// -*- C++ -*-
//
// Copyright (C) 1998, 1999, 2000, 2002  Los Alamos National Laboratory,
// Copyright (C) 1998, 1999, 2000, 2002  CodeSourcery, LLC
//
// This file is part of FreePOOMA.
//
// FreePOOMA is free software; you can redistribute it and/or modify it
// under the terms of the Expat license.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the Expat
// license for more details.
//
// You should have received a copy of the Expat license along with
// FreePOOMA; see the file LICENSE.
//

//-----------------------------------------------------------------------------
// Classes: 
//   BrickBase<Dim> 
//     - Base class for Brick-like engines
//   BrickViewBase<Dim> 
//     - Base class for BrickView-like engines
//   DoubleSliceHelper<Dim,Dim2,BaseDim>
//     - Helper functions for initializing slice-of-slice BrickViewBases
//-----------------------------------------------------------------------------

#ifndef POOMA_ENGINE_BRICKBASE_H
#define POOMA_ENGINE_BRICKBASE_H

/** @file
 * @ingroup Engine
 * @brief
 * Base classes for Brick- & BrickView-like engines. 
 * Encapsulates domain, stride, and subsetting operations.
 */

//-----------------------------------------------------------------------------
// Includes:
//-----------------------------------------------------------------------------

#include "Domain/Interval.h"
#include "Domain/Range.h"
#include "Domain/Loc.h"
#include "Domain/SliceRange.h"
#include "Domain/SliceInterval.h"
#include "Layout/DomainLayout.h"
#include "Utilities/PAssert.h"


///////////////////////////////////////////////////////////////////////////////

//-----------------------------------------------------------------------------
// Forward declarations:
//-----------------------------------------------------------------------------

namespace Pooma {

template <int Dim> class BrickViewBase;


/**
 * Template metaprogram for offset calculations.  Helps Loc<Dim>
 * indexing performance a bit.
 */

template <int pos, int max>
struct OffsetCalc
{
	template <class Dom>
	static inline int apply(const Dom& dom, const int *strides_m)
	{
		return dom[pos].first()*strides_m[pos]
			+ OffsetCalc<pos+1,max>::apply(dom, strides_m);
	}
};

template <int end>
struct OffsetCalc<end, end>
{
	template <class Dom>
	static inline int apply(const Dom& dom, const int *strides_m)
	{
		return dom[end].first()*strides_m[end];
	}
};

template <>
struct OffsetCalc<1, 0>
{
	template <class Dom>
	static inline int apply(const Dom& dom, const int *strides_m)
	{
		return 0;
	}
};


/**
 * BrickBase<Dim> is the base-class for engines that have a Brick-like 
 * interpretation of a block of data; i.e. whose data is stored in a single
 * contiguous block of memory that is interpreted as a Dim-dimensional
 * "brick" using Fortran storage conventions.
 *
 * BrickBase caches a copy of the original strides and can zero and restore
 * the strides array at will. These operations are used by CompressibleBricks
 * and are only enabled if the BrickBase constructor specifies that the
 * compressible flag is true.
 *
 * The template parameters are:
 *   - Dim, the dimension of the Brick domain.
 *
 * BrickBase encapsulates the strides, domain, and the calculation
 * of indexing offsets brick-like engines.
 *
 * BrickViewBase<Dim>, which is defined below, is used to represent 
 * subsets of a data block described by a BrickBase. 
 * BrickViewBase serves as the base class for brick-view types of engines.
 */

template <int Dim>
class BrickBase 
{
public:

  //============================================================
  // Exported typedefs and constants
  //============================================================

  typedef Interval<Dim>  Domain_t;
  typedef DomainLayout<Dim> Layout_t;
  
  typedef BrickBase<Dim> This_t;
  
  enum { dimensions = Dim   };
  enum { brick      = true  };
  enum { zeroBased  = false };

  //============================================================
  // Constructors and Factory Methods
  //============================================================

  /// Default constructor. Creates an uninitialized BrickBase
  /// (i.e. empty domain, uninitialized strides, etc.)
  
  explicit BrickBase(bool compressible = false)
    : compressibleBase_m(compressible)
  { }

  /// Initialize the BrickBase with an Interval describing the Dim-
  /// dimensional domain to be indexed.
  
  explicit BrickBase(const Domain_t &dom, bool compressible = false)
    : layout_m(dom), compressibleBase_m(compressible)
  {
    init();
  }

  /// Initialize the BrickBase with a Node describing the Dim-
  /// dimensional domain to be indexed.
  
  explicit BrickBase(const Node<Domain_t> &node, bool compressible = false)
    : layout_m(node), compressibleBase_m(compressible)
  {
    init();
  }

  /// Initialize the BrickBase with a Layout describing the Dim-
  /// dimensional domain to be indexed.
  
  explicit BrickBase(const Layout_t &layout, bool compressible = false)
    : layout_m(layout), compressibleBase_m(compressible)
  {
    init();
  }

  // Copy constructor - Use compiler generated version.

  //============================================================
  // Destructor
  //============================================================

  ~BrickBase() {}

  //============================================================
  // Assignment operators - Use compiler generated version.
  //============================================================

  //============================================================
  // Accessor functions:
  //============================================================

  /// Return the domain.

  inline const Domain_t &domain() const { return layout_m.domain(); }

  /// Return the layout.

  inline const Layout_t &layout() const { return layout_m; }

  /// Return the strides array.
  
  inline const int *strides() const { return &strides_m[0]; }
  
  /// Return the true strides array.
  
  inline const int *originalStrides() const { return &ostrides_m[0]; }

  /// Return the first index value for the specified dimension.

  int first(int i) const { return layout_m.domain()[i].first(); }

  /// Used by constructors
  
  bool compressibleBase() const { return compressibleBase_m; }

// protected: // these should probably be protected, but we use them
              // publicly in testing.

  
  /// @name Offset Calculations
  /// These are provided for Domains and for sets of ints. 
  ///
  /// Only BrickBase provides the offset0 & offsetC versions. 
  /// offset0 treats the input domain as a zero-based domain giving 
  /// offsets from the beginning in each dimension, rather than as 
  /// points in the Brick's underlying domain. offsetC is similar, 
  /// except that it multiplies the first offset by strides_m[0].
  /// This is only for use in CompressibleBrick.
  ///
  /// The "int" versions do not assert that their dimensionality is
  /// correct. The inheriting class should make these checks.
  /// These are fairly short and there are a lot of them so we define
  /// them in the class body for simplicity.

  //@{

  template <class Domain>
  inline int offset(const Domain &dom) const
  { return off_m + offset0(dom); }
  
  template <class Domain>
  inline int offset0(const Domain &dom) const
  {
    CTAssert(Domain::dimensions == Dim);
    return dom[0].first() + OffsetCalc<1, Dim-1>::apply(dom, strides_m);
  }
  
  template <class Domain>
  inline int offsetC(const Domain &dom) const
  {
    CTAssert(Domain::dimensions == Dim);
    return OffsetCalc<0, Dim-1>::apply(dom, strides_m);
  }

  inline int offset() const
  { return off_m; }
  
  inline int baseOffset() const
  { return off_m; }
  
  inline int offset(int i0) const
  { return off_m + i0; }
  
  inline int offset(int i0, int i1) const
  { return off_m + i0 + i1*strides_m[1]; }
  
  inline int offset(int i0, int i1, int i2) const
  { return off_m + i0 + i1*strides_m[1] + i2*strides_m[2]; }

  inline int offset(int i0, int i1, int i2, int i3) const
  { return off_m + i0 + i1*strides_m[1] + i2*strides_m[2] + i3*strides_m[3]; }
                    
  inline int offset(int i0, int i1, int i2, int i3, int i4) const
  { return off_m + i0 + i1*strides_m[1] + i2*strides_m[2] + i3*strides_m[3]
                      + i4*strides_m[4]; }
                    
  inline int offset(int i0, int i1, int i2, int i3, int i4, int i5) const
  { return off_m + i0 + i1*strides_m[1] + i2*strides_m[2] + i3*strides_m[3]
                      + i4*strides_m[4] + i5*strides_m[5]; }
                    
  inline int offset(int i0, int i1, int i2, int i3, int i4, int i5, int i6) 
  const
  { return off_m + i0 + i1*strides_m[1] + i2*strides_m[2] + i3*strides_m[3] 
                      + i4*strides_m[4] + i5*strides_m[5] + i6*strides_m[6]; }
                    
  inline int offset0(int i0) const
  { return i0; }
  
  inline int offset0(int i0, int i1) const
  { return i0 + i1*strides_m[1]; }
  
  inline int offset0(int i0, int i1, int i2) const
  { return i0 + i1*strides_m[1] + i2*strides_m[2]; }
  
  inline int offset0(int i0, int i1, int i2, int i3) const
  { return i0 + i1*strides_m[1] + i2*strides_m[2] + i3*strides_m[3]; }
  
  inline int offset0(int i0, int i1, int i2, int i3, int i4) const
  { return i0 + i1*strides_m[1] + i2*strides_m[2] + i3*strides_m[3]
              + i4*strides_m[4]; }
  
  inline int offset0(int i0, int i1, int i2, int i3, int i4, int i5) const
  { return i0 + i1*strides_m[1] + i2*strides_m[2] + i3*strides_m[3]
              + i4*strides_m[4] + i5*strides_m[5]; }
  
  inline int offset0(int i0, int i1, int i2, int i3, int i4, int i5, int i6) 
  const
  { return i0 + i1*strides_m[1] + i2*strides_m[2] + i3*strides_m[3]
              + i4*strides_m[4] + i5*strides_m[5] + i6*strides_m[6]; }
    
  inline int offsetC(int i0) const 
  { return i0*strides_m[0]; }
  
  inline int offsetC(int i0, int i1) const 
  { return i0*strides_m[0] + i1*strides_m[1]; }
  
  inline int offsetC(int i0, int i1, int i2) const
  { return i0*strides_m[0] + i1*strides_m[1] + i2*strides_m[2]; }

  inline int offsetC(int i0, int i1, int i2, int i3) const
  { return i0*strides_m[0] + i1*strides_m[1] + i2*strides_m[2] 
         + i3*strides_m[3]; }
                    
  inline int offsetC(int i0, int i1, int i2, int i3, int i4) const
  { return i0*strides_m[0] + i1*strides_m[1] + i2*strides_m[2] 
         + i3*strides_m[3] + i4*strides_m[4]; }
                    
  inline int offsetC(int i0, int i1, int i2, int i3, int i4, int i5) const
  { return i0*strides_m[0] + i1*strides_m[1] + i2*strides_m[2] 
         + i3*strides_m[3] + i4*strides_m[4] + i5*strides_m[5]; }
                    
  inline int offsetC(int i0, int i1, int i2, int i3, int i4, int i5, int i6) 
  const
  { return i0*strides_m[0] + i1*strides_m[1] + i2*strides_m[2] 
         + i3*strides_m[3] + i4*strides_m[4] + i5*strides_m[5] 
         + i6*strides_m[6]; }

  //@}
  
protected:

  /// Initialization of strides and offset from the domain.

  void init()
  {
    strides_m[0] = 1;
    ostrides_m[0] = 1;
    off_m        = -domain()[0].first();
    for (int d = 1; d < Dim; ++d) {
      strides_m[d]  = strides_m[d-1]*domain()[d-1].length();
      ostrides_m[d] = strides_m[d];
      off_m        -= domain()[d].first()*strides_m[d];
    }
  }

  //============================================================
  // Mutator functions:
  //============================================================

  // These are used to modify the strides array. They are only 
  // used by compressible engines and will assert if called
  // on a noncompressible engine.
  
  void zeroStrides() 
  { for (int d = 0; d < Dim; ++d) strides_m[d] = 0; }
  
  void restoreStrides()
  { for (int d = 0; d < Dim; ++d) strides_m[d] = ostrides_m[d]; }
  
  //============================================================
  // Protected data
  //============================================================

  /// Layout.

  Layout_t layout_m;

  /// Strides through actual data block when stepping in different dimensions.
  /// We keep two copies - strides_m is used by the offset calculations. 
  /// If we are compressible, then when compressed, these will all be 
  /// set to zero. 
  
  int strides_m[Dim];

  int ostrides_m[Dim];
  
  /// Offset due to non-zero first elements.

  int off_m;
  
  /// Flag indicating whether or not the stride compressions routines
  /// are callable. This is only set by the constructor and cannot be
  /// read or modified. It could be integrated into the type (as another
  /// template parameter), but that would complicate certain operations,
  /// and the benefit would probably not be worth the cost in complexity.
  
  bool compressibleBase_m;
};



/**
 * BrickViewBase<Dim> is the base-class for engines that are "views"
 * into brick-like engines.
 *
 * The template parameters are:
 *   - Dim, the logical dimension of the view
 *   - BaseDim, the dimension of the underlying BrickBase being viewed
 *
 * If Dim < BaseDim, then the view is called "sliced".
 * 
 * BrickViewBase encapsulates the calculations of the strides, domains,
 * and indexing offsets for these complicated views. 
 *
 * The general template specifies the "sliced" view. We specialize below
 * for Dim == BaseDim, which avoids the complications of slicing.
 */

template <int Dim>
class BrickViewBase
{
public:
  
  //============================================================
  // Exported typedefs and constants
  //============================================================

  enum { dimensions = Dim  };
  enum { zeroBased  = true };
  
  typedef Interval<Dim>  Domain_t;
  typedef DomainLayout<Dim> Layout_t;
  
  //============================================================
  // Constructors
  //============================================================

  /// Default constructor - creates an uninitialized BrickViewBase 
  /// (which has an empty domain with uninitilazed
  /// strides, etc.).
  
  BrickViewBase() {}

  // Copy constructor - Use compiler generated one.

  typedef BrickViewBase<Dim> This_t;

  //@{

  /// This is a special copy constructor that can change the
  /// compressibility flag. Useful, for example, when constructing
  /// a brick-view of a compressible-brick-view.
  
  BrickViewBase(const This_t &bvbase, bool compressible)
  {
    *this = bvbase;
    compressibleBase_m = compressible;
    if (!compressible) restoreStrides();
  }

  BrickViewBase(const BrickBase<Dim> &base, bool compressible)
  {
    *this = BrickViewBase<Dim>(base, base.domain());
    compressibleBase_m = compressible;
    if (!compressible) restoreStrides();
  }

  //@}

  /// @name Subsetting Constructors.
  /// This is where the real complexity lives. There are two categories
  /// of views:
  /// - Slices (Dim < BaseDim):
  ///   These can be created either by subsetting a BrickBase with a 
  ///   SliceInterval or a SliceRange, by subsetting a BrickViewBase
  ///   with any rectangular domain (Interval, Range, SliceInterval,
  ///   SliceRange).
  /// - Non-slices (Dim == BaseDim):
  ///   These are created by subsetting a BrickBase with an Interval
  ///   or Range, or by subsetting another non-sliced view with an
  ///   Interval or Range. These are handled by partial specializations
  ///   that fully specialize on the stride since that allows us to
  ///   only define the useful constructors.
  //@{

  BrickViewBase(const BrickBase<Dim> &bbase, const Interval<Dim> &dom)
    : domain_m(Pooma::NoInit()),
      baseOffset_m(bbase.offset()),
      compressibleBase_m(bbase.compressibleBase())
  {
    viewInit(bbase, dom);
  }

  BrickViewBase(const BrickBase<Dim> &bbase, const Range<Dim> &dom)
    : domain_m(Pooma::NoInit()),
      baseOffset_m(bbase.offset()),
      compressibleBase_m(bbase.compressibleBase())
  {
    viewInit(bbase, dom);
  }

  BrickViewBase(const This_t &bvbase, const Interval<Dim> &domain)
    : domain_m(Pooma::NoInit()),  
      baseOffset_m(bvbase.baseOffset()), 
      compressibleBase_m(bvbase.compressibleBase())
  {
    viewInit(bvbase, domain);
  }

  BrickViewBase(const This_t &bvbase, const Range<Dim> &domain)
    : domain_m(Pooma::NoInit()),  
      baseOffset_m(bvbase.baseOffset()), 
      compressibleBase_m(bvbase.compressibleBase())
  {
    viewInit(bvbase, domain);
  }

  template<int BaseDim>
  BrickViewBase(const BrickBase<BaseDim> &bbase,
		const SliceRange<BaseDim,Dim> &dom)
    : domain_m(Pooma::NoInit()), 
      baseOffset_m(bbase.offset()),
      compressibleBase_m(bbase.compressibleBase())
  {
    sliceInit(bbase.originalStrides(), dom);
  }

  template<int BaseDim>
  BrickViewBase(const BrickBase<BaseDim> &bbase,
		const SliceInterval<BaseDim,Dim> &dom)
    : domain_m(Pooma::NoInit()), 
      baseOffset_m(bbase.offset()),
      compressibleBase_m(bbase.compressibleBase())
  {
    sliceInit(bbase.originalStrides(), SliceRange<BaseDim,Dim>(dom));
  }

  template <int BaseDim>
  BrickViewBase(const BrickViewBase<BaseDim> &bvbase,
                const SliceRange<BaseDim,Dim> &dom)
  : domain_m(Pooma::NoInit()),
    baseOffset_m(bvbase.baseOffset())
  {
    sliceInit(bvbase, dom);
  }

  template <int BaseDim>
  BrickViewBase(const BrickViewBase<BaseDim> &bvbase, 
                const SliceInterval<BaseDim,Dim> &dom)
  : domain_m(Pooma::NoInit()),
    baseOffset_m(bvbase.baseOffset())
  {
    sliceInit(bvbase, SliceRange<BaseDim,Dim>(dom));
  }

  //@}

  //============================================================
  // Destructor
  //============================================================

  ~BrickViewBase() {}

  //============================================================
  // Assignment operators - Use compiler generated one.
  //============================================================

  //============================================================
  // Accessor functions:
  //============================================================

  /// Return our logical domain:

  inline const Domain_t &domain() const { return domain_m; }

  /// Return our layout:

  inline Layout_t layout() const { return Layout_t(domain_m); }

  /// Return the strides array.

  inline const int *strides() const { return &strides_m[0]; }
  
  /// Return the true strides array.
  
  inline const int *originalStrides() const { return &ostrides_m[0]; }

  /// Return the first index value for the specified dimension.
  /// (Always zero since views are zero-based).
  
  inline int first(int) const { return 0; }

  /// Used by constructors
  
  bool compressibleBase() const { return compressibleBase_m; }

  /// Base offset - offset from beginning of underlying Brick's beginning.
  
  int baseOffset() const { return baseOffset_m; }
  
  /// @name Offset calculations
  //@{

  template <class Domain>
  inline int offset(const Domain &dom) const
  {
    CTAssert(Domain::dimensions == Dim);
    return OffsetCalc<0, Dim-1>::apply(dom, strides_m);
  }
  
  inline int offset(int i0) const 
  { return i0*strides_m[0]; }
  
  inline int offset(int i0, int i1) const 
  { return i0*strides_m[0] + i1*strides_m[1]; }
  
  inline int offset(int i0, int i1, int i2) const
  { return i0*strides_m[0] + i1*strides_m[1] + i2*strides_m[2]; }

  inline int offset(int i0, int i1, int i2, int i3) const
  { return i0*strides_m[0] + i1*strides_m[1] + i2*strides_m[2] 
         + i3*strides_m[3]; }
                    
  inline int offset(int i0, int i1, int i2, int i3, int i4) const
  { return i0*strides_m[0] + i1*strides_m[1] + i2*strides_m[2] 
         + i3*strides_m[3] + i4*strides_m[4]; }
                    
  inline int offset(int i0, int i1, int i2, int i3, int i4, int i5) const
  { return i0*strides_m[0] + i1*strides_m[1] + i2*strides_m[2] 
         + i3*strides_m[3] + i4*strides_m[4] + i5*strides_m[5]; }
                    
  inline int offset(int i0, int i1, int i2, int i3, int i4, int i5, int i6)
    const
  { return i0*strides_m[0] + i1*strides_m[1] + i2*strides_m[2] 
         + i3*strides_m[3] + i4*strides_m[4] + i5*strides_m[5] 
         + i6*strides_m[6]; }


  template <class Domain>
  inline int offsetU(const Domain &dom) const
  {
    CTAssert(Domain::dimensions == Dim);
    return dom[0].first() + OffsetCalc<1, Dim-1>::apply(dom, strides_m);
  }
  
  inline int offsetU(int i0) const 
  { return i0; }
  
  inline int offsetU(int i0, int i1) const 
  { return i0 + i1*strides_m[1]; }
  
  inline int offsetU(int i0, int i1, int i2) const
  { return i0 + i1*strides_m[1] + i2*strides_m[2]; }

  inline int offsetU(int i0, int i1, int i2, int i3) const
  { return i0 + i1*strides_m[1] + i2*strides_m[2] 
         + i3*strides_m[3]; }
                    
  inline int offsetU(int i0, int i1, int i2, int i3, int i4) const
  { return i0 + i1*strides_m[1] + i2*strides_m[2] 
         + i3*strides_m[3] + i4*strides_m[4]; }
                    
  inline int offsetU(int i0, int i1, int i2, int i3, int i4, int i5) const
  { return i0 + i1*strides_m[1] + i2*strides_m[2] 
         + i3*strides_m[3] + i4*strides_m[4] + i5*strides_m[5]; }
                    
  inline int offsetU(int i0, int i1, int i2, int i3, int i4, int i5, int i6)
    const
  { return i0 + i1*strides_m[1] + i2*strides_m[2] 
         + i3*strides_m[3] + i4*strides_m[4] + i5*strides_m[5] 
         + i6*strides_m[6]; }

  //@}
  
protected:

  //============================================================
  // Mutator functions:
  //============================================================

  // These are used to modify the strides array. They are only 
  // used by compressible engines and will assert if called
  // on a noncompressible engine.
  
  void zeroStrides() 
  { for (int d = 0; d < Dim; ++d) strides_m[d] = 0; }
  
  void restoreStrides() 
  { for (int d = 0; d < Dim; ++d) strides_m[d] = ostrides_m[d]; }
  
  //============================================================
  // Utility functions
  //============================================================

  //---------------------------------------------------------------------------
  //
  // sliceInit(const Interval<BaseDim>&,
  //           const SliceRange<BaseDim,Dim> &)
  //
  // Helper function used in taking a slice of a Brick or a Brick-like view.
  //---------------------------------------------------------------------------

  template<int BaseDim>
  void sliceInit(const int *baseStrides,
		 const SliceRange<BaseDim,Dim> &dom);

  template<int BaseDim>
  void sliceInit(const BrickViewBase<BaseDim>&, const SliceRange<BaseDim, Dim>&);

  void viewInit(const This_t &, const Range<Dim> &domain);

  void viewInit(const BrickBase<Dim> &bbase, const Range<Dim> &domain);

  void viewInit(const This_t &bvbase, const Interval<Dim> &domain)
  {
    for (int d = 0; d < Dim; ++d)
    {
      domain_m[d]   = Interval<1>(domain[d].length());
      strides_m[d]  = bvbase.ostrides_m[d];
      ostrides_m[d] = strides_m[d];
      baseOffset_m += domain[d].first() * bvbase.ostrides_m[d];
    }
  }

  void viewInit(const BrickBase<Dim> &bbase, const Interval<Dim> &domain)
  {
    for (int d = 0; d < Dim; ++d)
    {
      domain_m[d]   = Interval<1>(domain[d].length());
      strides_m[d]  = bbase.originalStrides()[d];
      ostrides_m[d] = strides_m[d];
      baseOffset_m  += domain[d].first() * bbase.originalStrides()[d];
    }
  }


  //============================================================
  // Data
  //============================================================

  // Domain for this engine:
  // (Somewhat wasteful since we are zero-based and only need
  // the length. However, we may want to return references to this
  // object for efficiency reasons, and thus we can't create it
  // on the fly when someone asks.)
  
  Domain_t domain_m;

  // Strides through actual data block when stepping in each dimension.
  // We keep two copies - strides_m is used by the offset calculations. 
  // If we are compressible, then when compressed, these will all be 
  // set to zero. 
  
  int strides_m[Dim];

  int ostrides_m[Dim];
    
  // Base offset - offset of beginning of view from underlying Brick's
  // beginning.
  
  int baseOffset_m;
  
  // Compressibility flag - see BrickBase comments.
  
  bool compressibleBase_m;
};


template <int Dim>
void
BrickViewBase<Dim>::
viewInit(const This_t &bvbase, const Range<Dim> &domain)
{
  // Compute the strides and domain.
    
  for (int d = 0; d < Dim; ++d)
    {
      domain_m[d]   = Interval<1>(domain[d].length());
      strides_m[d]  = bvbase.ostrides_m[d] * domain[d].stride();
      baseOffset_m += domain[d].first() * bvbase.ostrides_m[d];
    }

  for (int d = 0; d < Dim; ++d)
    ostrides_m[d] = strides_m[d];
}

template <int Dim>
void
BrickViewBase<Dim>::
viewInit(const BrickBase<Dim> &bbase, const Range<Dim> &domain)
{
  for (int d = 0; d < Dim; ++d)
  {
    domain_m[d]   = Interval<1>(domain[d].length());
    strides_m[d]  = bbase.originalStrides()[d] * domain[d].stride();
    ostrides_m[d] = strides_m[d];
    baseOffset_m  += domain[d].first() * bbase.originalStrides()[d];
  }
}

template <int Dim>
template<int BaseDim>
void
BrickViewBase<Dim>::
sliceInit(const int *baseStrides,
	  const SliceRange<BaseDim,Dim> &dom)
{
  typedef typename SliceRange<BaseDim,Dim>::TotalDomain_t TotalDomain_t;
  const TotalDomain_t &domain = dom.totalDomain();

  // compute the dt-dim base stride along in the loop
  int d = 0;    
  for (int dt = 0; dt < BaseDim; ++dt)
  {
    if (!dom.ignorable(dt))
    {
      PBoundAssert(d < Dim);
      domain_m[d]  = Interval<1>(domain[dt].length());
      strides_m[d] = baseStrides[dt] * domain[dt].stride();
      ++d;
    }

    baseOffset_m += domain[dt].first() * baseStrides[dt];
  }

  PBoundAssert(d == Dim);

  for (int d = 0; d < Dim; ++d)
    ostrides_m[d] = strides_m[d];
}

template <int Dim>
template<int BaseDim>
void
BrickViewBase<Dim>::
sliceInit(const BrickViewBase<BaseDim>& bvbase, const SliceRange<BaseDim, Dim>& dom)
{  
  // Compute the domain and strides.
  // The domain is an Interval with the length of each component
  // equal to the length of the corrsponding domain in the SliceRange.
  // The strides are calculated by multiplying the strides in 
  // the non-ignorable directions of the engine being viewed, by the
  // strides in the SliceRange that is doing the viewing. Since we must
  // skip over the ignorable dimensions in the engine being viewed,
  // we write this as a loop over the viewed engine's dimensions and
  // just do nothing for the ignorable ones.

  typedef typename SliceRange<BaseDim,Dim>::TotalDomain_t TotalDomain_t;
  const TotalDomain_t &totDomain = dom.totalDomain();

  int d, dt;
  for (dt = 0, d = 0; dt < BaseDim; ++dt)
    {
      if (!dom.ignorable(dt))
        {
          PBoundAssert(d < Dim);
          domain_m[d]  = Interval<1>(totDomain[dt].length());
          strides_m[d] = bvbase.originalStrides()[dt] * totDomain[dt].stride();
          ++d;
        }
      baseOffset_m += totDomain[dt].first() * bvbase.originalStrides()[dt];
    }
  PBoundAssert(d == Dim);

  for (int d = 0; d < Dim; ++d)
    ostrides_m[d] = strides_m[d];
}


} // namespace Pooma



///////////////////////////////////////////////////////////////////////////////

#endif // POOMA_ENGINE_BRICKBASE_H

// ACL:rcsinfo
// ----------------------------------------------------------------------
// $RCSfile: BrickBase.h,v $   $Author: richi $
// $Revision: 1.25 $   $Date: 2005/04/13 20:42:04 $
// ----------------------------------------------------------------------
// ACL:rcsinfo
