// -*- C++ -*-
//
// Copyright (C) 2003, 2004  Richard Guenther
//
// This file is part of FreePOOMA.
//
// FreePOOMA is free software; you can redistribute it and/or modify it
// under the terms of the Expat license.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the Expat
// license for more details.
//
// You should have received a copy of the Expat license along with
// FreePOOMA; see the file LICENSE.
//

/** @file
 * @ingroup Evaluator
 * @brief
 * PatchCode functionality for applying a functor to a patch.  Like
 * ScalarCode, but with manual control over the loops.
 *
 * Note that with PatchCode there is no OpenMP parallelism until you
 * add that manually to your functors.
 */

#ifndef POOMA_EVALUATOR_PATCHCODE_H
#define POOMA_EVALUATOR_PATCHCODE_H

//////////////////////////////////////////////////////////////////////

//-----------------------------------------------------------------------------
// Overview: 
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// Typedefs:
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// Includes:
//-----------------------------------------------------------------------------

#include "Functions/MultiArg.h"
#include "Evaluator/MultiArgEvaluator.h"
#include "Evaluator/ScalarCodeInfo.h"

//-----------------------------------------------------------------------------
// Forward Declarations:
//-----------------------------------------------------------------------------


template<class Function, int Dim>
struct EvaluatePatch
{
  EvaluatePatch()
  {
  }

  EvaluatePatch(const Function &function, const Interval<Dim> &domain)
    : function_m(function), domain_m(domain)
  {
  }

  template<class A1>
  void operator()(MultiArg1<A1> &m)
  {
    function_m(m.a1_m, domain_m);
  }

  template<class A1, class A2>
  void operator()(MultiArg2<A1, A2> &m)
  {
    function_m(m.a1_m, m.a2_m, domain_m);
  }

  template<class A1, class A2, class A3>
  void operator()(MultiArg3<A1, A2, A3> &m)
  {
    function_m(m.a1_m, m.a2_m, m.a3_m, domain_m);
  }

  template<class A1, class A2, class A3, class A4>
  void operator()(MultiArg4<A1, A2, A3, A4> &m)
  {
    function_m(m.a1_m, m.a2_m, m.a3_m, m.a4_m, domain_m);
  }

  template<class A1, class A2, class A3, class A4, class A5>
  void operator()(MultiArg5<A1, A2, A3, A4, A5> &m)
  {
    function_m(m.a1_m, m.a2_m, m.a3_m, m.a4_m, m.a5_m, domain_m);
  }

  template<class A1, class A2, class A3, class A4, class A5, class A6>
  void operator()(MultiArg6<A1, A2, A3, A4, A5, A6> &m)
  {
    function_m(m.a1_m, m.a2_m, m.a3_m, m.a4_m, m.a5_m, m.a6_m, domain_m);
  }

  template<class A1, class A2, class A3, class A4, class A5, class A6, class A7>
  void operator()(MultiArg7<A1, A2, A3, A4, A5, A6, A7> &m)
  {
    function_m(m.a1_m, m.a2_m, m.a3_m, m.a4_m, m.a5_m, m.a6_m, m.a7_m, domain_m);
  }

  const Function& function_m;
  const Interval<Dim>& domain_m;
};



/**
 * PatchCode is a Stencil like operation that allows for more than one
 * field to be operated on. Generally the functor is a local (set of)
 * function(s) which could be described as
 *
 *   (f1..fM) = op(fM+1..fN)
 *
 * where fM+1 to fN are input fields read from and f1 to fM are output
 * fields written to (this distinction nor its ordering is strictly
 * required, but both will result in the least possible surprises).
 */

template<class Function>
struct PatchCode
{
  PatchCode()
  {
  }

  PatchCode(const Function &function)
    : function_m(function)
  {
  }

  /// Constructor to allow PatchCode being used as RelationFunctor

  template <class LHS>
  PatchCode(const PatchCode<Function>& sc, const LHS&)
    : function_m(sc.function_m)
  {
  }

  template<class F>
  static inline bool checkValidity(const F& f, WrappedInt<false>)
  {
    // Arrays are ok always.
    return true;
  }

  template<class F>
  static inline bool checkValidity(const F& f, WrappedInt<true>)
  {
    // We cannot handle fields with arbitrary centering size
    // due to physicalDomain() giving potentially wrong answers.
    // We don't handle materials either.
    return f.centeringSize() == 1 && f.numMaterials() == 1;
  }

  /// @name Evaluators
  /// Evaluate the PatchCode functor on the fields f1 to fN using the
  /// specified evaluation domain. Note that views of the evaluation domain
  /// are taken of every field, so domains of the fields should be strictly
  /// conforming (in fact, passing views to these operators is a bug unless
  /// you really know what you are doing).
  ///
  /// The evaluation domain defaults to the physical domain of
  /// the first field which should usually be (on of) the left hand side(s).
  /// If you want the functor to operate on a different domain use the
  /// operators with the explicit specified evaluation domain.
  //@{

  template<class F1>
  void operator()(const F1 &f1, const Interval<F1::dimensions> &evalDom) const
  {
    PAssert(checkValidity(f1, WrappedInt<F1::hasRelations>()));
    MultiArg1<F1> multiArg(f1);
    EvaluatePatch<Function, F1::dimensions> kernel(function_m, evalDom);
    MultiArgEvaluator<MainEvaluatorTag>::
      evaluate(multiArg, function_m, evalDom, kernel);
  }

  template<class F1>
  inline void operator()(const F1 &f1) const
  {
    (*this)(f1, f1.physicalDomain());
  }


  template<class F1, class F2>
  void operator()(const F1 &f1, const Interval<F1::dimensions> &evalDom,
                  const F2 &f2) const
  {
    PAssert(checkValidity(f1, WrappedInt<F1::hasRelations>()));
    MultiArg2<F1, F2> multiArg(f1, f2);
    EvaluatePatch<Function, F1::dimensions> kernel(function_m, evalDom);
    MultiArgEvaluator<MainEvaluatorTag>::
      evaluate(multiArg, function_m, evalDom, kernel);
  }

  template<class F1, class F2>
  inline void operator()(const F1 &f1, const F2 &f2) const
  {
    (*this)(f1, f1.physicalDomain(), f2);
  }


  template<class F1, class F2, class F3>
  void operator()(const F1 &f1, const Interval<F1::dimensions> &evalDom,
                  const F2 &f2, const F3 &f3) const
  {
    PAssert(checkValidity(f1, WrappedInt<F1::hasRelations>()));
    MultiArg3<F1, F2, F3> multiArg(f1, f2, f3);
    EvaluatePatch<Function, F1::dimensions> kernel(function_m, evalDom);
    MultiArgEvaluator<MainEvaluatorTag>::
      evaluate(multiArg, function_m, evalDom, kernel);
  }

  template<class F1, class F2, class F3>
  inline void operator()(const F1 &f1, const F2 &f2, const F3 &f3) const
  {
    (*this)(f1, f1.physicalDomain(), f2, f3);
  }


  template<class F1, class F2, class F3, class F4>
  void operator()(const F1 &f1, const Interval<F1::dimensions> &evalDom,
		  const F2 &f2, const F3 &f3, const F4 &f4) const
  {
    PAssert(checkValidity(f1, WrappedInt<F1::hasRelations>()));
    MultiArg4<F1, F2, F3, F4> multiArg(f1, f2, f3, f4);
    EvaluatePatch<Function, F1::dimensions> kernel(function_m, evalDom);
    MultiArgEvaluator<MainEvaluatorTag>::
      evaluate(multiArg, function_m, evalDom, kernel);
  }

  template<class F1, class F2, class F3, class F4>
  inline void operator()(const F1 &f1, const F2 &f2, const F3 &f3, const F4 &f4) const
  {
    (*this)(f1, f1.physicalDomain(), f2, f3, f4);
  }


  template<class F1, class F2, class F3, class F4, class F5>
  void operator()(const F1 &f1, const Interval<F1::dimensions> &evalDom,
		  const F2 &f2, const F3 &f3, const F4 &f4, const F5 &f5) const
  {
    PAssert(checkValidity(f1, WrappedInt<F1::hasRelations>()));
    MultiArg5<F1, F2, F3, F4, F5> multiArg(f1, f2, f3, f4, f5);
    EvaluatePatch<Function, F1::dimensions> kernel(function_m, evalDom);
    MultiArgEvaluator<MainEvaluatorTag>::
      evaluate(multiArg, function_m, evalDom, kernel);
  }

  template<class F1, class F2, class F3, class F4, class F5>
  inline void operator()(const F1 &f1, const F2 &f2, const F3 &f3, const F4 &f4,
			 const F5 &f5) const
  {
    (*this)(f1, f1.physicalDomain(), f2, f3, f4, f5);
  }


  template<class F1, class F2, class F3, class F4, class F5, class F6>
  void operator()(const F1 &f1, const Interval<F1::dimensions> &evalDom,
		  const F2 &f2, const F3 &f3, const F4 &f4, const F5 &f5,
		  const F6 &f6) const
  {
    PAssert(checkValidity(f1, WrappedInt<F1::hasRelations>()));
    MultiArg6<F1, F2, F3, F4, F5, F6> multiArg(f1, f2, f3, f4, f5, f6);
    EvaluatePatch<Function, F1::dimensions> kernel(function_m, evalDom);
    MultiArgEvaluator<MainEvaluatorTag>::
      evaluate(multiArg, function_m, evalDom, kernel);
  }

  template<class F1, class F2, class F3, class F4, class F5, class F6>
  inline void operator()(const F1 &f1, const F2 &f2, const F3 &f3, const F4 &f4,
			 const F5 &f5, const F6 &f6) const
  {
    (*this)(f1, f1.physicalDomain(), f2, f3, f4, f5, f6);
  }


  template<class F1, class F2, class F3, class F4, class F5, class F6, class F7>
  void operator()(const F1 &f1, const Interval<F1::dimensions> &evalDom,
		  const F2 &f2, const F3 &f3, const F4 &f4,
		  const F5 &f5, const F6 &f6, const F7 &f7) const
  {
    PAssert(checkValidity(f1, WrappedInt<F1::hasRelations>()));
    MultiArg7<F1, F2, F3, F4, F5, F6, F7> multiArg(f1, f2, f3, f4, f5, f6, f7);
    EvaluatePatch<Function, F1::dimensions> kernel(function_m, evalDom);
    MultiArgEvaluator<MainEvaluatorTag>::
      evaluate(multiArg, function_m, evalDom, kernel);
  }

  template<class F1, class F2, class F3, class F4, class F5, class F6, class F7>
  inline void operator()(const F1 &f1, const F2 &f2, const F3 &f3, const F4 &f4,
			 const F5 &f5, const F6 &f6, const F7 &f7) const
  {
    (*this)(f1, f1.physicalDomain(), f2, f3, f4, f5, f6, f7);
  }

  //@}

  Function function_m;
};



//////////////////////////////////////////////////////////////////////

#endif     // POOMA_EVALUATOR_PATCHCODE_H

// ACL:rcsinfo
// ----------------------------------------------------------------------
// $RCSfile: PatchCode.h,v $   $Author: richi $
// $Revision: 1.1 $   $Date: 2004/12/22 13:43:32 $
// ----------------------------------------------------------------------
// ACL:rcsinfo
