// -*- C++ -*-
//
// Copyright (C) 2003, 2004  Richard Guenther
//
// This file is part of FreePOOMA.
//
// FreePOOMA is free software; you can redistribute it and/or modify it
// under the terms of the Expat license.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the Expat
// license for more details.
//
// You should have received a copy of the Expat license along with
// FreePOOMA; see the file LICENSE.
//

/** @file
 * @ingroup IO
 * @brief
 * Traits classes for mapping POOMA types to HDF5 objects and wrappers
 * for some HDF5 functions with POOMA style domain specifiers.
 */

#ifndef POOMA_IO_HDF5_H
#define POOMA_IO_HDF5_H

#include "PoomaConfiguration.h"

#include <hdf5.h>
#include <cstring>
#include "Pooma/Domains.h"
#include "Pooma/BrickArrays.h"
#include "Pooma/Tiny.h"



/**
 * Traits class specialized for types that qualify as primitive.
 * The only method provided is
 *   hid_t type();
 * which returns a HDF5 type representing the type in memory.
 */

template <class T>
struct HDF5TypeTraits
{
  static ErrorType type();
};



/**
 * Upon construction, object gets updated,
 * upon destruction after write, object gets marked dirty.
 * HDF5Object may create temporary objects for accessing storage of object.
 *
 * The default implementation handles primitive types as scalars.
 */

template <class T>
struct HDF5Object
{
  typedef T Object_t;
  enum { usesTemporary = false };

  HDF5Object(const Object_t& o) : o_m(o) {}
  HDF5Object(Object_t& o, hid_t) : o_m(o) {}

  hid_t type() const { return HDF5TypeTraits<T>::type(); }
  hid_t dataspace() const { return H5Screate(H5S_SCALAR); }
  void* write() const { return &(const_cast<Object_t&>(o_m)); }
  const void* read() const { return &o_m; }

  const Object_t& o_m;
};


/// Overloads for H5Screate_simple for Pooma domain objects.
/// 
/// HDF5 prototype is
///   hid_t H5Screate_simple(int rank, const hsize_t *dims, const hsize_t *maxdims);

template <int Dim>
hid_t H5Screate_simple(const Interval<Dim> &I)
{
  hsize_t dims[Dim];
  for (int i=0; i<Dim; i++)
    dims[Dim-1-i] = I[i].size();
  return H5Screate_simple(Dim, dims, NULL);
}

template <int Dim>
void H5Sget_simple_extent_domain(hid_t, Interval<Dim>&); /// umm...

template <int Dim>
void H5Sget_select_hyperslab(hid_t, Interval<Dim>&); /// umm... SliceRange<>?


/// Overloads for H5Sselect_hyperslab for Pooma domain objects.
/// Note that the domain of the dataspace is assumed to be zero-based.
/// Also we can't support Grid<> style selections with HDF5 hyperslabs.
///
/// HDF5 prototype is
///   herr_t H5Sselect_hyperslab(hid_t space_id, H5S_seloper_t op,
///                              const hssize_t *start, const hsize_t *stride,
///                              const hsize_t *count, const hsize_t *block);

template <int Dim>
herr_t H5Sselect_hyperslab(hid_t space_id, H5S_seloper_t op,
			   const Interval<Dim> &I)
{
  hssize_t sel_start[Dim];
  hsize_t sel_count[Dim];
  for (int i=0; i<Dim; i++) {
    sel_start[Dim-1-i] = I[i].first();
    sel_count[Dim-1-i] = I[i].size();
  }
  return H5Sselect_hyperslab(space_id, op,
			     sel_start, NULL, sel_count, NULL);
}

template <int Dim>
herr_t H5Sselect_hyperslab(hid_t space_id, H5S_seloper_t op,
			   const Range<Dim> &R)
{
  hssize_t sel_start[Dim];
  hsize_t sel_count[Dim];
  hsize_t sel_stride[Dim];
  for (int i=0; i<Dim; i++) {
    sel_start[Dim-1-i] = R[i].first();
    sel_count[Dim-1-i] = R[i].size();
    sel_stride[Dim-1-i] = R[i].stride();
  }
  return H5Sselect_hyperslab(space_id, op,
			     sel_start, sel_stride, sel_count, NULL);
}

template <int TotalDim, int SliceDim>
herr_t H5Sselect_hyperslab(hid_t space_id, H5S_seloper_t op,
			   const SliceInterval<TotalDim, SliceDim> &I);

template <int TotalDim, int SliceDim>
herr_t H5Sselect_hyperslab(hid_t space_id, H5S_seloper_t op,
			   const SliceRange<TotalDim, SliceDim> &R);


// Get all the traits specializations.

#include "IO/HDF5.cpp"


#endif // defined POOMA_IO_HDF5_H
