// -*- C++ -*-
//
// Copyright (C) 1998, 1999, 2000, 2002  Los Alamos National Laboratory,
// Copyright (C) 1998, 1999, 2000, 2002  CodeSourcery, LLC
//
// This file is part of FreePOOMA.
//
// FreePOOMA is free software; you can redistribute it and/or modify it
// under the terms of the Expat license.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the Expat
// license for more details.
//
// You should have received a copy of the Expat license along with
// FreePOOMA; see the file LICENSE.
//

//-----------------------------------------------------------------------------
// Classes: 
//   UniformGridLayout<Dim>
//   UniformGridLayoutView<Dim, Dim2>
//   UniformTag
//   MultiPatchLayoutTraits<UniformTag,Dim>
//-----------------------------------------------------------------------------

#ifndef POOMA_LAYOUT_UNIFORMGRIDLAYOUT_H
#define POOMA_LAYOUT_UNIFORMGRIDLAYOUT_H

/** @file
 * @ingroup Layout
 * @brief
 *   UniformGridLayout<Dim>
 *     - Layout class that breaks Dim-dimensional domain into equal
 *       sized sub-domains arranged in a Dim-dimensional grid.
 *   UniformGridLayoutView<Dim, Dim2>
 *     - view of a UniformGridLayout
 *   UniformTag
 *     - tag used to specialize MultiPatchLayoutTraits
 *   MultiPatchLayoutTraits<UniformTag,Dim>
 *     - traits class used by MultiPatch-engine to determine layout type.
 */

//-----------------------------------------------------------------------------
// Includes:
//-----------------------------------------------------------------------------

#include "Layout/INode.h"
#include "Layout/TouchesConstruct.h"
#include "Layout/GuardLayers.h"
#include "Domain/Interval.h"
#include "Domain/Contains.h"
#include "Domain/Intersect.h"
#include "Domain/NewDomain.h"
#include "Domain/SliceRange.h"
#include "Partition/UniformGridPartition.h"
#include "Partition/ContextMapper.h"
#include "Utilities/DerefIterator.h"
#include "Utilities/ViewIndexer.h"
#include "Utilities/Observable.h"
#include "Utilities/Observer.h"
#include "Utilities/RefCountedPtr.h"
#include "Utilities/RefCounted.h"
#include "Utilities/Unique.h"
#include "Utilities/PAssert.h"

#include "Layout/LayoutBase.h"

#include <vector>
#include <iosfwd>


///////////////////////////////////////////////////////////////////////////////
// namespace Pooma {

//============================================================
// Forward declaration
//============================================================

template <int Dim> class UniformGridLayout;
template <int Dim, int Dim2> class UniformGridLayoutView;

//-----------------------------------------------------------------------------
// Full Description:
// UniformTag
//
// Tag class.
//-----------------------------------------------------------------------------

struct UniformTag { };


/**
 * Specialization of MultiPatchLayoutTraits for UniformGridLayout.
 */

template <int Dim>
struct MultiPatchLayoutTraits<UniformTag,Dim>
{
  typedef UniformGridLayout<Dim> Layout_t;
  
  template <int ViewDim>
  struct View
  {
    typedef UniformGridLayoutView<ViewDim,Dim> Layout_t;
  };
};


/**
 * This is the actual data for the UniformGridLayout class, which is
 * simply a wrapper that holds a reference counted instance of this
 * data class.
 */

template <int Dim>
class UniformGridLayoutData 
 : public LayoutBaseData<Dim>, 
   public RefCounted, 
   public Observable<UniformGridLayoutData<Dim> >
{
public:
  //============================================================
  // Typedefs and enumerations
  //============================================================

  // General public typedefs.

  typedef Interval<Dim>           Domain_t;
  typedef Interval<Dim>           BaseDomain_t;
  typedef int                     Context_t;
  typedef Unique::Value_t         ID_t;
  typedef Node<Domain_t>          Value_t;
  typedef std::vector<Value_t *>  List_t;           // for convenience
  typedef GuardLayers<Dim>        GuardLayers_t;    // for convenience
  
  typedef typename LayoutBaseData<Dim>::GCFillInfo GCFillInfo_t;
  
  typedef typename std::vector<GCFillInfo_t>::const_iterator FillIterator_t;
  
  // Enumerations.

  enum { dimensions = Dim };
  enum { repartitionEvent = 1 };
  enum { dynamic = false };


  //============================================================
  // Constructors
  //============================================================

  // Default version creates an empty domain.
  
  UniformGridLayoutData();
  
  // All initialization is done via the following constructor.
  // Originally we had several constructors, but once we went
  // to not storing a partitioner object, it was easier to just
  // have the Layout construct the partitioner appropriately
  // and pass it to the LayoutData constructor.
  
  template <class Partitioner>
  UniformGridLayoutData(const Domain_t &gdom, 
			const Partitioner &gpar,
			const ContextMapper<Dim> & cmap );


  //============================================================
  // Destructor
  //============================================================

  // This is only called when all references to this data go away, in
  // which case we need to delete our nodes. The Observable destructor
  // will broadcast messages up to all observers of the Layout.

  ~UniformGridLayoutData()
  {
    typename List_t::iterator a;
    for (a = this->all_m.begin(); a != this->all_m.end(); ++a)
      delete (*a);
  }


  // Find all subdomains that touch on a given domain, and insert the
  // intersection of these subdomains into the given output iterator.
  // Return the number of touching elements. This version of touches
  // can build either pointers or objects.

  template <class OtherDomain, class OutIter, class ConstructTag>
  int touches(const OtherDomain &d, OutIter o, const ConstructTag &ctag) const;

  template <class OtherDomain, class OutIter, class ConstructTag>
  int touchesLocal(const OtherDomain &d, 
		   OutIter o, 
		   const ConstructTag &ctag) const;

  // Find all Remote subdomains that touch on a given domain, and insert the
  // intersection of these subdomains into the given output iterator.
  // Return the number of touching elements. This version of touches
  // can build either pointers or objects.

  template <class OtherDomain, class OutIter, class ConstructTag>
  int touchesRemote(const OtherDomain &d, 
		    OutIter o, 
		    const ConstructTag &ctag) const;

  template <class OtherDomain, class OutIter, class ConstructTag>
  int touchesAlloc(const OtherDomain &d, OutIter o, 
                   const ConstructTag &ctag) const;

  template <class OtherDomain, class OutIter, class ConstructTag>
  int touchesAllocLocal(const OtherDomain &d, OutIter o, 
			const ConstructTag &ctag) const;

  template <class OtherDomain, class OutIter, class ConstructTag>
  int touchesAllocRemote(const OtherDomain &d, OutIter o, 
			 const ConstructTag &ctag) const;

  //private:

  friend class UniformGridLayout<Dim>;

  //============================================================
  // Accessors
  //============================================================


  // These are deferred to UniformGridLayoutData in order to avoid a
  // bunch of indirections in their implementation.

  // Accessors for getting the global ID of the patch containing a
  // particular element.

  int globalID(const Loc<Dim> &loc) const;
  int globalID(int) const;
  int globalID(int,int) const;
  int globalID(int,int,int) const;
  int globalID(int,int,int,int) const;
  int globalID(int,int,int,int,int) const;
  int globalID(int,int,int,int,int,int) const;
  int globalID(int,int,int,int,int,int,int) const;
  
  // Mutators used for partitioning.

  template <class Partitioner>
  void partition(const Partitioner &, const ContextMapper<Dim>& cmap);

  template <class Partitioner>
  bool repartition(const Partitioner &,const ContextMapper<Dim>&);
  
  // This function calculates the cached guard-cell filling information.
  
  void calcGCFillList();
  
  
  //============================================================
  // Data
  //============================================================

  // The stride array for indexing into the 1-D list of blocks;
  // i.e. in 2D, block (i,j) is element (i + j * blockstride_m[1]) in
  // the list.
  
  int blockstride_m[Dim];

  // The patch size. Could be stored in an interval, but use an array
  // since this info is zero-based.
  
  int blocksizes_m[Dim];

  // The domain of the "brick" of patches stored in the all_m list.
  
  Interval<Dim> allDomain_m;
};


/**
 * UniformGridLayout is a Layout class that breaks an N-dimensional
 * Interval into equal sized sub-domains arranged in an N-dimensional
 * grid.
 *
 * This is an alternative to the more general Layout class that should
 * perform somewhat faster since subdomains can be found
 * arithmetically, rather than via a search.  It is only able to
 * represent grid-like layouts, however.
 *
 * To construct a UniformGridLayout, you can do any of the following:
 *   -# provide a global domain, and let the UniformGridLayout perform
 *      its default partitioning by just using one single block;
 *   -# provide a global domain, a Loc with the number of blocks to
 *      use along each dimension
 *   -# provide a global domain and a UniformGridPartition object.
 *      
 * Alternatively, you can use the default UniformGridLayout
 * constructor, and call the 'initialize' method later with the same
 * possible set of arguments.
 *
 * You can also specify internal and external guard layers for the
 * domains. See the comments in UniformGridLayout below.
 */

template <int Dim>
class UniformGridLayout : public LayoutBase<Dim,UniformGridLayoutData<Dim> >,
                          public Observable<UniformGridLayout<Dim> >,
                          public Observer<UniformGridLayoutData<Dim> >
{
public:

  //============================================================
  // Typedefs and enumerations
  //============================================================

  // General public typedefs.
  
  typedef UniformGridLayoutData<Dim>           LayoutData_t; // for convenience
  typedef typename LayoutData_t::Domain_t      Domain_t;
  typedef typename LayoutData_t::BaseDomain_t  BaseDomain_t;
  typedef typename LayoutData_t::Context_t     Context_t;
  typedef typename LayoutData_t::ID_t          ID_t;
  typedef typename LayoutData_t::Value_t       Value_t;
  typedef typename LayoutData_t::List_t        List_t; // for convenience
  typedef UniformGridLayout<Dim>               This_t; // for convenience
  typedef Observable<This_t>                   Observable_t;

  // Iterator through nodes. Basically the same as the vector iterator
  // except it dereferences automatically.
  
  typedef DerefIterator<Value_t>               iterator;
  typedef ConstDerefIterator<Value_t>          const_iterator;
  
  // Iterator through guard-cell-fill requests. 
  
  typedef typename LayoutData_t::GCFillInfo_t GCFillInfo_t;
  
  typedef typename 
    std::vector<GCFillInfo_t>::const_iterator FillIterator_t;
  
  typedef GuardLayers<Dim> GuardLayers_t;

  // Enumerations.

  enum { dimensions = Dim };
  enum { repartitionEvent = LayoutData_t::repartitionEvent };
  enum { dynamic = false };


  //============================================================
  // Constructors
  //============================================================

  /**
  // The default constructor does not initialize the layout.  In this
  // case, layout initialization must be completed with the
  // 'initialize' method before the layout can be used.  A default
  // layout has an empty global domain, and empty subdomain lists.
  
  // This is also the only constructor that doesn't demand either
  // ReplicatedTag or DistributedTag
  */
  
  UniformGridLayout();

  // Distributed versions


  UniformGridLayout(const Domain_t &,
		    const DistributedTag &);

  UniformGridLayout(const Domain_t &, 
                    const GuardLayers_t &,
		    const DistributedTag &);

  UniformGridLayout(const Domain_t &, 
                    const Loc<Dim> &,
		    const DistributedTag &);

  UniformGridLayout(const Domain_t &, 
                    const Loc<Dim> &, 
                    const GuardLayers_t &,
		    const DistributedTag &);

  UniformGridLayout(const Domain_t &, 
                    const Loc<Dim> &, 
                    const GuardLayers_t &, 
                    const GuardLayers_t &,
		    const DistributedTag &);
  //========================= Replicated versions ===================

  UniformGridLayout(const Domain_t &,
		    const ReplicatedTag &);

  UniformGridLayout(const Domain_t &, 
                    const GuardLayers_t &,
		    const ReplicatedTag &);

  UniformGridLayout(const Domain_t &, 
                    const Loc<Dim> &,
		    const ReplicatedTag &);

  UniformGridLayout(const Domain_t &, 
                    const Loc<Dim> &, 
                    const GuardLayers_t &,
		    const ReplicatedTag &);

  UniformGridLayout(const Domain_t &, 
                    const Loc<Dim> &, 
                    const GuardLayers_t &, 
                    const GuardLayers_t &,
		    const ReplicatedTag &);

  // Domain + partition constructor. 
  // In this case, the partitioner must be a UniformGridPartition
  // object. The partitioner must know about guard cells, so we don't
  // specify them separately here.
  
  template <class Partitioner>
  UniformGridLayout(const Domain_t &,
		    const Partitioner &,
		    const ContextMapper<Dim> & );

  template <class Partitioner>
  UniformGridLayout(const Domain_t &, 
                    const Partitioner &,
		    const DistributedTag &);

  template <class Partitioner>
  UniformGridLayout(const Domain_t &, 
                    const Partitioner &,
		    const ReplicatedTag &);

  // Copy constructor & assignment operator
  // Shallow copies with reference counting.
  
  UniformGridLayout(const This_t &);
  
  This_t &operator=(const This_t &);

  // I think we need a version of this here ... just forward to
  // the RefCountedPtr.
  
  // makeOwnCopy???


  //============================================================
  // Destructor
  //============================================================

  // The actual data will be cleaned up by the LayoutData_t destructor
  // if all references to the data go away.  If any Observers remain,
  // they will be notified by the Observable destructor.
  
  inline ~UniformGridLayout() 
  { 
    this->pdata_m->detach(*this);
  }


  //============================================================
  // Initialize methods
  //============================================================

  // Initialize a layout with nothing else but a global domain.  In
  // this case, a default partitioner will be used, the
  // UniformGridPartition object, which will try to make a grid with
  // one block.  
  
  void initialize(const Domain_t &,
		  const DistributedTag &);
  
  void initialize(const Domain_t &,
		  const ReplicatedTag &);
  
  void initialize(const Domain_t &, 
		  const GuardLayers_t &,
		  const DistributedTag &);
  
  void initialize(const Domain_t &, 
		  const GuardLayers_t &,
		  const ReplicatedTag &);
  
  void initialize(const Domain_t &, 
		  const Loc<Dim> &,
		  const DistributedTag &);

  void initialize(const Domain_t &, 
		  const Loc<Dim> &,
		  const ReplicatedTag &);

  void initialize(const Domain_t &, 
		  const Loc<Dim> &, 
		  const GuardLayers_t &,
		  const DistributedTag & );

  void initialize(const Domain_t &, 
		  const Loc<Dim> &, 
		  const GuardLayers_t &,
		  const ReplicatedTag & );

  void initialize(const Domain_t &, 
                  const Loc<Dim> &, 
                  const GuardLayers_t &, 
                  const GuardLayers_t &,
		  const DistributedTag &);

  void initialize(const Domain_t &, 
                  const Loc<Dim> &, 
                  const GuardLayers_t &, 
                  const GuardLayers_t &,
		  const ReplicatedTag &);

  // Domain + partition initializer.
  // In this case, the partitioner must be a UniformGridPartition
  // object.
  
  template <class Partitioner>
  void initialize(const Domain_t &, 
		  const Partitioner &,
		  const DistributedTag &);

  template <class Partitioner>
  void initialize(const Domain_t &, 
		  const Partitioner &,
		  const ReplicatedTag &);
  
  // Domain + partition + mapper  initializer.
  // In this case, the partitioner must be a UniformGridPartition
  // object.
  
  template <class Partitioner>
  void initialize(const Domain_t &, 
		  const Partitioner &,
		  const ContextMapper<Dim> &);
 
  // Used by the I/O or data management system to initialize the layout based
  // on detailed state information previously stored. Since this is specialized
  // for the I/O system, no trailing tag is used. 

  void initialize(const Domain_t& idom,
		  const List_t& nodes,
		  const Loc<Dim>& blocks,
		  bool hasIG, bool hasEG,
		  const GuardLayers_t& ig,
		  const GuardLayers_t& eg);
  //============================================================
  // Accessors
  //============================================================




  //============================================================
  // Observer methods
  //============================================================

  // Respond to events generated by the LayoutData_t.
  // These are just passed on to our observers.
  // this needs to be here, and not in the base class

  virtual void notify(LayoutData_t &d, const ObserverEvent &event)
  {
    // We should only get this message from our LayoutData_t object
    PAssert(&d == this->pdata_m.rawPointer());
    Observable_t::notify(event);
  }
    

  //============================================================
  // Output
  //============================================================
    
  // Print a UniformGridLayout on an output stream

  template <class Ostream>
  void print(Ostream &ostr) const;


#if !POOMA_NO_TEMPLATE_FRIENDS
  // can't seem to get specialized template friends working...
  //private:

  template <int Dim1, int Dim2>
  friend class UniformGridLayoutView;

#endif
  
  //============================================================
  // Data
  //===========================================================

  // UniformGridLayout stores its data in a LayoutBase::RefCounted class to
  // simplify memory management.
  
  friend class UniformGridLayoutData<Dim>;

};



/**
 * This is the actual data for the UniformGridLayoutView class, which is
 * simply a wrapper that holds a reference counted instance of this
 * data class.
 */

template <int Dim, int Dim2>
class UniformGridLayoutViewData 
  : public LayoutBaseViewData<Dim, Dim2, UniformGridLayout<Dim2> >,
    public RefCounted
{
public:

  typedef UniformGridLayout<Dim2>              Layout_t;
  typedef UniformGridLayoutView<Dim, Dim2>     ViewLayout_t;
  typedef LayoutBaseViewData<Dim,Dim2,Layout_t> Base_t;
  typedef Interval<Dim>                        Domain_t;
  typedef Range<Dim2>                          BaseDomain_t;
  typedef int                                  Context_t;
  typedef Unique::Value_t                      ID_t;

  typedef typename Layout_t::Domain_t          AllocatedDomain_t;
  typedef ViewIndexer<Dim,Dim2>                Indexer_t;

  typedef Node<Domain_t,AllocatedDomain_t>     Value_t;
  typedef std::vector<Value_t *>               List_t;        // convenience
  typedef GuardLayers<Dim>                     GuardLayers_t; // convenience


  // Enumerations.

  enum { dim = Dim };
  enum { dim2 = Dim2};

  //============================================================
  // Constructors
  //============================================================

  UniformGridLayoutViewData() { };
  
  template <class DT>
  inline 
  UniformGridLayoutViewData(const Layout_t &layout, const Domain<Dim, DT> &dom)
  : LayoutBaseViewData<Dim,Dim2,UniformGridLayout<Dim2> >(layout,dom)
  { 
  }

  template <class DT>
  inline 
  UniformGridLayoutViewData(const Layout_t &layout, const SliceDomain<DT> &dom)
  :LayoutBaseViewData<Dim,Dim2,UniformGridLayout<Dim2> >(layout,dom)
  {  
  }

  template <class DT>
  UniformGridLayoutViewData(const ViewLayout_t &layout, 
                            const Domain<Dim, DT> &dom)
  : LayoutBaseViewData<Dim,Dim2,UniformGridLayout<Dim2> >(
					      layout.pdata_m->layout_m,
					      layout,
					      layout.pdata_m->indexer_m, 
					      dom,
					      layout.internalGuards(),
					      layout.externalGuards())
  {
  }

  template <int OrigDim, class DT>
  UniformGridLayoutViewData(const UniformGridLayoutView<OrigDim,Dim2> &layout, 
                            const SliceDomain<DT> &dom)
  : LayoutBaseViewData<Dim,Dim2,UniformGridLayout<Dim2> >(
			     layout.pdata_m->layout_m,
			     layout,
			     Indexer_t(layout.pdata_m->indexer_m,dom),
			     dom)
  { 
  }

  // Destructor

  ~UniformGridLayoutViewData() 
  {
    typename List_t::iterator a;
    for (a = this->all_m.begin(); a != this->all_m.end(); ++a)
      delete (*a);
  }

};


/**
 * UniformGridLayoutView is a Layout class that provides a view of an
 * existing UniformGridLayout object. Dim is the logical dimension of
 * the layout. Dim2 is the dimension of the UniformGridLayout
 * contained within.
 *
 * To construct a UniformGridLayoutView, you need an existing
 * UniformGridLayout or a UniformGridLayoutView and the subdomain that
 * is being viewed. This class does not have a useful default
 * constructor since it is based on an existing UniformGridLayout.
 *
 * Once created, UniformGridLayoutView has the same interface as
 * Layout (see Layout.h). It also provides this extra interface:
 *
 * int globalID(const Loc<Dim> &pos) : return the globalID
 *    of the node that contains the point.
 */

template <int Dim, int Dim2>
class UniformGridLayoutView
  : public LayoutBaseView<Dim, Dim2, UniformGridLayoutViewData<Dim,Dim2> >
{
public:
  //============================================================
  // Typedefs and enumerations
  //============================================================

  // Enumerations.

  enum { dimensions = Dim };

  enum { dim = Dim };
  enum { dim2 = Dim2};
  // General public typedefs.
  
  typedef UniformGridLayoutViewData<Dim, Dim2>      LayoutData_t; 

  typedef typename LayoutData_t::Domain_t           Domain_t;
  typedef typename LayoutData_t::BaseDomain_t       BaseDomain_t;
  typedef typename LayoutData_t::Context_t          Context_t;
  typedef typename LayoutData_t::ID_t               ID_t;
  
  typedef typename LayoutData_t::Layout_t           Layout_t;
  typedef typename LayoutData_t::AllocatedDomain_t  AllocatedDomain_t;
  typedef typename LayoutData_t::Value_t            Value_t;
  
  typedef typename LayoutData_t::List_t             List_t;
  typedef typename LayoutData_t::Indexer_t          Indexer_t;
  typedef typename LayoutData_t::GuardLayers_t      GuardLayers_t;
    
  typedef UniformGridLayoutView<Dim, Dim2>          This_t;      // convenience
  typedef UniformGridLayoutView<Dim, Dim2>          ViewLayout_t;// convenience
  typedef LayoutBaseView<Dim,Dim2,LayoutData_t>     Base_t;

  // Iterator through nodes. Basically the same as the vector iterator
  // except it dereferences automatically.
  
  typedef DerefIterator<Value_t>                    iterator;
  typedef ConstDerefIterator<Value_t>               const_iterator;


  //============================================================
  // Constructors
  //============================================================

  // Default constructor. Final initialization should be done with
  // the assignment operator. 
  
  UniformGridLayoutView()
  : Base_t(new LayoutData_t())
  { }
  
  // Constructor building a UniformGridLayoutView from a
  // UniformGridLayout and a non-slice domain like an Interval<Dim> or
  // Range<Dim>.
  
  template <class DT>
  UniformGridLayoutView(const Layout_t &layout, const Domain<Dim2, DT> &dom)
  : LayoutBaseView<Dim,Dim2,UniformGridLayoutViewData<Dim,Dim2> >
    (new UniformGridLayoutViewData<Dim,Dim2>(layout,dom))
  { }

  // Constructor building a UniformGridLayoutView from a
  // UniformGridLayout and a slice domain like a
  // SliceInterval<Dim2,Dim> or SliceRange<Dim2,Dim>.
  
  template <class DT>
  UniformGridLayoutView(const Layout_t &layout, const SliceDomain<DT> &dom)
  : LayoutBaseView<Dim,Dim2,UniformGridLayoutViewData<Dim,Dim2> >
    (new UniformGridLayoutViewData<Dim,Dim2>(layout,dom))
  { }
  
  // Constructor building a UniformGridLayoutView from another
  // UniformGridLayoutView and a non-slice domain like an
  // Interval<Dim> or Range<Dim>.
  
  template <class DT>
  UniformGridLayoutView(const ViewLayout_t &layout, const Domain<Dim, DT> &dom)
  : LayoutBaseView<Dim,Dim2,UniformGridLayoutViewData<Dim,Dim2> >
    (new UniformGridLayoutViewData<Dim,Dim2>(layout,dom))
  { }

  // Constructor building a UniformGridLayoutView from another
  // UniformGridLayoutView and a slice domain like a
  // SliceInterval<Dim2,Dim> or SliceRange<Dim2,Dim>.
  
  template <int OldViewDim, class DT>
  UniformGridLayoutView(const UniformGridLayoutView<OldViewDim, Dim2> &layout, 
                        const SliceDomain<DT> &dom)
  : LayoutBaseView<Dim,Dim2,UniformGridLayoutViewData<Dim,Dim2> >
    (new UniformGridLayoutViewData<Dim,Dim2>(layout,dom))
  { }

  // Copy constructor & assignment operator
  // Shallow copies with reference counting.
  
  inline UniformGridLayoutView(const This_t &model) 
    : LayoutBaseView<Dim,
                     Dim2,
                     UniformGridLayoutViewData<Dim,Dim2> >(model.pdata_m)
  { }
  
  inline This_t &operator=(const This_t &model) 
  {
    if (this != &model)
      {
        this->pdata_m = model.pdata_m;
      }
    return *this;
  }

  // I think we need a version of this here ... just forward to
  // the RefCountedPtr.
  
  // makeOwnCopy???



  //============================================================
  // Destructor
  //============================================================

  // The actual data will be cleaned up by the UniformGridLayoutData
  // destructor if all references to the data go away, so there is
  // nothing to do here.
   
  inline ~UniformGridLayoutView() 
  { }

  //============================================================
  // Output
  //============================================================
    
  // Print a UniformGridLayoutView on an output stream

  template <class Ostream>
  void print(Ostream &ostr) const;
  
#if !POOMA_NO_TEMPLATE_FRIENDS

  //protected:

  template <int OtherDim, int OtherDim2>
  friend class UniformGridLayoutView;
  
  template <int OtherDim, int OtherDim2>
  friend class UniformGridLayoutViewData;

#endif

  //============================================================
  // Data
  //============================================================

  // The data is stored in a LayoutBaseView::RefCounted class to simplify 
  // memory management. (LayoutViews have shallow copy semantics and they
  // own copies of the underlying Layout as well as their own sets of
  // Node lists, which must be cleaned up when all views are destroyed.)

};


//============================================================
// NewDomain1 traits classes for UniformGridLayout and
// UniformGridLayoutView
//============================================================

//-----------------------------------------------------------------------------
//
// This is so an array can be initialized with a UniformGridLayout.
//
//-----------------------------------------------------------------------------

template <int Dim>
struct NewDomain1<UniformGridLayout<Dim> >
{
  typedef UniformGridLayout<Dim> &Type_t;

  inline static Type_t combine(const UniformGridLayout<Dim> &a)
    {
      return const_cast<Type_t>(a);
    }
};

//-----------------------------------------------------------------------------
//
// This is so an array can be initialized with a UniformGridLayoutView.
//
//-----------------------------------------------------------------------------

template <int Dim, int Dim2>
struct NewDomain1<UniformGridLayoutView<Dim, Dim2> >
{
  typedef UniformGridLayoutView<Dim, Dim2> &Type_t;

  inline static Type_t combine(const UniformGridLayoutView<Dim, Dim2> &a)
    {
      return const_cast<Type_t>(a);
    }
};

//-----------------------------------------------------------------------------
//
// ostream inserters for UniformGridLayout and UniformGridLayoutView:
//
//-----------------------------------------------------------------------------

template <int Dim>
std::ostream &operator<<(std::ostream &ostr, 
                         const UniformGridLayout<Dim> &layout)
{
  layout.print(ostr);
  return ostr;
}

template <int Dim, int Dim2>
std::ostream &operator<<(std::ostream &ostr, 
                         const UniformGridLayoutView<Dim, Dim2> &layout)
{
  layout.print(ostr);
  return ostr;
}


// } // namespace POOMA

#include "Layout/UniformGridLayout.cpp"

#endif // POOMA_LAYOUT_UNIFORMGRIDLAYOUT_H

// ACL:rcsinfo
// ----------------------------------------------------------------------
// $RCSfile: UniformGridLayout.h,v $   $Author: richi $
// $Revision: 1.90 $   $Date: 2005/04/11 12:08:52 $
// ----------------------------------------------------------------------
// ACL:rcsinfo
