// -*- C++ -*-
//
// Copyright (C) 2002, 2004  Richard Guenther
//
// This file is part of FreePOOMA.
//
// FreePOOMA is free software; you can redistribute it and/or modify it
// under the terms of the Expat license.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the Expat
// license for more details.
//
// You should have received a copy of the Expat license along with
// FreePOOMA; see the file LICENSE.
//

#ifndef POOMA_TINY_ONE_H
#define POOMA_TINY_ONE_H

//-----------------------------------------------------------------------------
// Class: One
//-----------------------------------------------------------------------------

/** @file
 * @ingroup Tiny
 * @brief
 * A numeric class for a number that is always one.
 */

//-----------------------------------------------------------------------------
// Typedefs:
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// Includes:
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// Forward Declarations:
//-----------------------------------------------------------------------------

template<class T> struct Zero;
template<class T, class Op> struct UnaryReturn;
template<class T1, class T2, class Op> struct BinaryReturn;


/// The point of this class is to be a number of type T which is known at
/// compile time to be one.
///
/// This is reflected in the operations like
/// multiplication and division that use it.  If you return a One object,
/// the compiler should be able to make better optimizations than if you
/// just return one.
/// One is templated on type T, to represent a one object of type T.
/// Type T is required to be constructable from an integer one.

template<class T>
struct One
{
  /// If you need to convert to an object of type T,
  /// just build one from one.  This will be used in 
  /// the cases where the operators below don't match.

  operator T() const { return T(1); }

  bool operator==(const One<T>&) const { return true; }

  /// Give it empty ctors and assignment operators
  /// to try and keep purify happy.

  One() {}
  One(const One<T>&) {}
  One<T>& operator=(const One<T>&) { return *this; }
};

/// The point of this class is to be a number of type T which is known at
/// compile time to be minus one.
///
/// This is reflected in the operations like
/// multiplication and division that use it.  If you return a MinusOne object,
/// the compiler should be able to make better optimizations than if you
/// just return minus one.
/// MinusOne is templated on type T, to represent a one object of type T.
/// Type T is required to be constructable from an integer minus one.

template<class T>
struct MinusOne
{
  /// If you need to convert to an object of type T,
  /// just build one from minus one.  This will be used in 
  /// the cases where the operators below don't match.

  operator T() const { return T(-1); }

  bool operator==(const MinusOne<T>&) const { return true; }

  /// Give it empty ctors and assignment operators
  /// to try and keep purify happy.

  MinusOne() {}
  MinusOne(const One<T>&) {}
  MinusOne<T>& operator=(const MinusOne<T>&) { return *this; }
};


//-----------------------------------------------------------------------------
//
// Operators using One and/or MinusOne.
//
//-----------------------------------------------------------------------------

//
// Unary minus operators.
//

template<class T>
  inline MinusOne<T>  operator-(One<T>) { return MinusOne<T>(); }

template<class T>
  inline One<T>  operator-(MinusOne<T>) { return One<T>(); }


//
// Binary multiply/divide operations involving One<T>/MinusOne<T>.
//

template<class T> 
  inline T  operator*(One<T>, const T& val)   { return val; }

template<class T> 
  inline T  operator*(const T& val, One<T>)   { return val; }

template<class T> 
  inline One<T>  operator*(One<T>, One<T>)    { return One<T>(); }

template<class T>
  inline One<T>  operator/(One<T>, One<T>)    { return One<T>(); }

template<class T> 
  inline T  operator/(const T& val, One<T>)   { return val; }

template<class T> 
  inline T  operator*(MinusOne<T>, const T& val)   { return -val; }

template<class T> 
  inline T  operator*(const T& val, MinusOne<T>)   { return -val; }

template<class T> 
  inline One<T>  operator*(MinusOne<T>, MinusOne<T>)    { return One<T>(); }

template<class T> 
  inline One<T>  operator/(MinusOne<T>, MinusOne<T>)    { return One<T>(); }

template<class T> 
  inline T  operator/(const T& val, MinusOne<T>)   { return -val; }

template<class T> 
  inline MinusOne<T> operator*(MinusOne<T>, One<T>)   { return MinusOne<T>(); }

template<class T> 
  inline MinusOne<T> operator*(One<T>, MinusOne<T>)   { return MinusOne<T>(); }

template<class T> 
  inline MinusOne<T> operator/(MinusOne<T>, One<T>)   { return MinusOne<T>(); }

template<class T> 
  inline MinusOne<T> operator/(One<T>, MinusOne<T>)   { return MinusOne<T>(); }

//
// Subtraction/addition of a One<T>/MinusOne<T> from a One<T>/MinusOne<T>
// which yields Zero<T>.
//

template<class T>
  inline Zero<T> operator-(One<T>, One<T>)    { return Zero<T>(); }

template<class T> 
  inline Zero<T> operator-(MinusOne<T>, MinusOne<T>)   { return Zero<T>(); }

template<class T> 
  inline Zero<T> operator+(One<T>, MinusOne<T>)   { return Zero<T>(); }

template<class T> 
  inline Zero<T> operator+(MinusOne<T>, One<T>)   { return Zero<T>(); }


template<class T>
  inline MinusOne<T> operator-(Zero<T>, One<T>)   { return MinusOne<T>(); }

template<class T>
  inline One<T> operator-(Zero<T>, MinusOne<T>)   { return One<T>(); }



//
// Trait classes so that expression templates will deal correctly
// with Ones/MinusOnes in specific operations.
//

// Unary minus

template<class T>
struct UnaryReturn< One<T> , OpUnaryMinus >
{
  typedef MinusOne<T> Type_t;
};

template<class T>
struct UnaryReturn< MinusOne<T> , OpUnaryMinus >
{
  typedef One<T> Type_t;
};

// Multiply

template<class T>
struct BinaryReturn< One<T> , T , OpMultiply >
{
  typedef T Type_t;
};

template<class T>
struct BinaryReturn< T, One<T> , OpMultiply >
{
  typedef T Type_t;
};

template<class T>
struct BinaryReturn< One<T>, One<T> , OpMultiply >
{
  typedef One<T> Type_t;
};

template<class T>
struct BinaryReturn< MinusOne<T> , T , OpMultiply >
{
  typedef T Type_t;
};

template<class T>
struct BinaryReturn< T, MinusOne<T> , OpMultiply >
{
  typedef T Type_t;
};

template<class T>
struct BinaryReturn< MinusOne<T>, MinusOne<T> , OpMultiply >
{
  typedef One<T> Type_t;
};

template<class T>
struct BinaryReturn< MinusOne<T>, One<T> , OpMultiply >
{
  typedef MinusOne<T> Type_t;
};

template<class T>
struct BinaryReturn< One<T>, MinusOne<T> , OpMultiply >
{
  typedef MinusOne<T> Type_t;
};

// Divide

template<class T>
struct BinaryReturn< T, One<T> , OpDivide >
{
  typedef T Type_t;
};

template<class T>
struct BinaryReturn< T, MinusOne<T> , OpDivide >
{
  typedef T Type_t;
};

template<class T>
struct BinaryReturn< MinusOne<T>, One<T> , OpDivide >
{
  typedef MinusOne<T> Type_t;
};

template<class T>
struct BinaryReturn< One<T>, MinusOne<T> , OpDivide >
{
  typedef MinusOne<T> Type_t;
};

// Subtract

template<class T>
struct BinaryReturn< One<T>, One<T> , OpSubtract >
{
  typedef Zero<T> Type_t;
};

template<class T>
struct BinaryReturn< MinusOne<T>, MinusOne<T> , OpSubtract >
{
  typedef Zero<T> Type_t;
};

template<class T>
struct BinaryReturn< Zero<T>, One<T>, OpSubtract >
{
  typedef MinusOne<T> Type_t;
};

template<class T>
struct BinaryReturn< Zero<T>, MinusOne<T>, OpSubtract >
{
  typedef One<T> Type_t;
};

// Add

template<class T>
struct BinaryReturn< One<T>, MinusOne<T> , OpAdd >
{
  typedef Zero<T> Type_t;
};

template<class T>
struct BinaryReturn< MinusOne<T>, One<T> , OpAdd >
{
  typedef Zero<T> Type_t;
};

// Equality

template<class T>
struct BinaryReturn< MinusOne<T>, MinusOne<T>, OpEQ>
{
  typedef bool Type_t;
};

template<class T>
struct BinaryReturn< One<T>, One<T>, OpEQ>
{
  typedef bool Type_t;
};


//
// Trait class definitios for the cases where the One<T>/MinusOne<T> gets
// converted to a T. Traits with just one specialized operand are
// ambigous to specialized operator versions.
//

template<class T1, class T2, class Op>
struct BinaryReturn< One<T1> , One<T2>, Op >
{
  typedef typename BinaryReturn<T1,T2,Op>::Type_t Type_t;
};

template<class T1, class T2, class Op>
struct BinaryReturn< MinusOne<T1> , MinusOne<T2>, Op >
{
  typedef typename BinaryReturn<T1,T2,Op>::Type_t Type_t;
};


//
// Let Zero, One and MinusOne play together for other operators.
//

template<class T1, class T2, class Op>
struct BinaryReturn< One<T1>, Zero<T2>, Op >
{
  typedef typename BinaryReturn<T1,T2,Op>::Type_t Type_t;
};

template<class T1, class T2, class Op>
struct BinaryReturn< Zero<T1>, One<T2>, Op >
{
  typedef typename BinaryReturn<T1,T2,Op>::Type_t Type_t;
};

template<class T1, class T2, class Op>
struct BinaryReturn< MinusOne<T1>, Zero<T2>, Op >
{
  typedef typename BinaryReturn<T1,T2,Op>::Type_t Type_t;
};

template<class T1, class T2, class Op>
struct BinaryReturn< Zero<T1>, MinusOne<T2>, Op >
{
  typedef typename BinaryReturn<T1,T2,Op>::Type_t Type_t;
};

template<class T1, class T2, class Op>
struct BinaryReturn< One<T1>, MinusOne<T2>, Op >
{
  typedef typename BinaryReturn<T1,T2,Op>::Type_t Type_t;
};

template<class T1, class T2, class Op>
struct BinaryReturn< MinusOne<T1>, One<T2>, Op >
{
  typedef typename BinaryReturn<T1,T2,Op>::Type_t Type_t;
};



#endif

// ACL:rcsinfo
// ----------------------------------------------------------------------
// $RCSfile: One.h,v $   $Author: richi $
// $Revision: 1.1 $   $Date: 2005/02/15 13:42:46 $
// ----------------------------------------------------------------------
// ACL:rcsinfo
